#!/usr/bin/env python
#
# This software is distributed with the MIT license:
# https://github.com/gkiar/clowdr/blob/master/LICENSE
#
# clowdr/controller/rerunner.py
# Created by Greg Kiar on 2018-06-11.
# Email: gkiar@mcin.ca

import os.path as op
import json
import sys
import os
import re

from clowdr import utils


def getTasks(provdir, runid, rerun_mode):
    runpath = utils.truepath(op.join(provdir, runid, 'clowdr'))

    files = os.listdir(runpath)
    r_all = re.compile('^.*task-([0-9]+)[.]json$')
    all_tasks = sorted([op.join(runpath, f)
                        for f in filter(r_all.match, files)])

    if rerun_mode == "all":
        return all_tasks

    all_ids = set([r_all.match(f).group(1)
                   for f in all_tasks
                   if r_all.match(f)])

    task_dict = {tid:task for (tid, task) in zip(sorted(all_ids), all_tasks)}

    r_complete = re.compile('^.*task-([0-9]+)-summary[.]json$')
    complete_ids = set([r_complete.match(f).group(1)
                        for f in files
                        if r_complete.match(f)])

    incomplete_ids = sorted(all_ids - complete_ids)

    if rerun_mode == "incomplete":
        return [task_dict[tid] for tid in incomplete_ids]

    complete_summaries = sorted([op.join(runpath, f)
                                 for f in filter(r_complete.match, files)])

    failed_ids = []
    for (tid, summary) in zip(sorted(complete_ids), complete_summaries):
        with open(summary, 'r') as fhandle:
            task_summary = json.load(fhandle)
            if task_summary["exitcode"]:
                failed_ids += [tid]

    return [task_dict[tid] for tid in failed_ids]



#!/usr/bin/env python
#
# This software is distributed with the MIT license:
# https://github.com/gkiar/clowdr/blob/master/LICENSE
#
# clowdr/endpoint/AWS.py
# Created by Greg Kiar on 2018-02-28.
# Email: gkiar@mcin.ca

from botocore.exceptions import *
import time
import boto3
import os.path as op
import json
import warnings
import os
import re

warnings.filterwarnings("ignore", message="numpy.dtype size changed")

import pandas as pd

from clowdr.endpoint.remote import Endpoint
from clowdr import __path__ as clowfile


clowfile = clowfile[0]


class AWS(Endpoint):
    # TODO: document

    def setCredentials(self, **kwargs):
        # TODO: document

        creds = pd.read_csv(self.credentials)
        self.access_key = creds['Access key ID'][0]
        self.secret_access = creds['Secret access key'][0]
        os.environ["AWS_ACCESS_KEY_ID"] = self.access_key
        os.environ["AWS_SECRET_ACCESS_KEY"] = self.secret_access

        if kwargs.get("region"):
            self.region = kwargs["region"]
        else:
            self.region = "us-east-1"

    def startSession(self):
        # TODO: document
        self.session = boto3.Session(aws_access_key_id=self.access_key,
                                     aws_secret_access_key=self.secret_access,
                                     region_name=self.region)
        self.iam = self.session.client("iam")
        self.ec2 = self.session.client("ec2")
        self.batch = self.session.client("batch")

    def configureIAM(self, **kwargs):
        # TODO: document
        template = op.join(op.realpath(clowfile), "templates",
                           "AWS", "userRoles.json")

        with open(template) as fhandle:
            roles = json.load(fhandle)
        basestr = "arn:aws:iam::aws:policy/service-role/"
        policy = {"batch": basestr + "AWSBatchServiceRole",
                  "ecs":   basestr + "AmazonEC2ContainerServiceforEC2Role"}

        for rolename in roles:
            role = roles[rolename]
            try:
                name = role["RoleName"]
                response = self.iam.get_role(RoleName=name)
                role["Arn"] = response["Role"]["Arn"]

            except ClientError as e:
                if e.response["Error"]["Code"] == "NoSuchEntity":
                    if kwargs.get("verbose"):
                        print("Role '{}' not found- creating.".format(name),
                              flush=True)
                    role["AssumeRolePolicyDocument"] = json.dumps(role["AssumeRolePolicyDocument"])
                    response = self.iam.create_role(**role)
                    role["Arn"] = response["Role"]["Arn"]
                    self.iam.create_instance_profile(InstanceProfileName=name)
                    self.iam.add_role_to_instance_profile(InstanceProfileName=name,
                                                          RoleName=name)
                    self.iam.attach_role_policy(RoleName=name,
                                                PolicyArn=policy[rolename])
                    roles[rolename] = role
            if kwargs.get("verbose"):
                print("Role ARN: {}".format(roles[rolename]["Arn"]),
                      flush=True)
        self.roles = roles

    def configureBatch(self, **kwargs):
        # TODO: document
        sg = [sg["GroupId"]
              for sg in self.ec2.describe_security_groups()["SecurityGroups"]
              if sg["GroupName"] == "default"]
        net = [nets["SubnetId"]
               for nets in self.ec2.describe_subnets()["Subnets"]]

        def waitUntilDone(name, status):
            while True:
                env = self.batch.describe_compute_environments(computeEnvironments=[name])
                stat = env["computeEnvironments"][0]["status"]
                if stat != status:
                    return

        template = op.join(op.realpath(clowfile), "templates",
                           "AWS", "computeEnvironment.json")
        with open(template) as fhandle:
            compute = json.load(fhandle)

        try:
            name = compute["computeEnvironmentName"]
            response = self.batch.describe_compute_environments(computeEnvironments=[name])
            if len(response["computeEnvironments"]):
                if (response["computeEnvironments"][0]["status"] != "VALID" or
                   response["computeEnvironments"][0]["state"] != "ENABLED"):
                    raise ClientError({"Error": {"Code": "InvalidEnvironment"}},
                                      "describe_compute_environments")
                else:
                    compute["computeEnvironmentArn"] = response["computeEnvironments"][0]["computeEnvironmentArn"]
            else:
                raise ClientError({"Error": {"Code": "NoSuchEntity"}},
                                  "describe_compute_environments")

        except ClientError as e:
            if e.response["Error"]["Code"] == "InvalidEnvironment":
                if kwargs.get("verbose"):
                    print("Environment '{}' invalid- deleting.".format(name),
                          flush=True)
                response = self.batch.update_compute_environment(computeEnvironment=name,
                                                                 state="DISABLED")
                waitUntilDone(name, "UPDATING")
                response = self.batch.delete_compute_environment(computeEnvironment=name)
                waitUntilDone(name, "DELETING")

            if (e.response["Error"]["Code"] == "NoSuchEntity" or
               e.response["Error"]["Code"] == "InvalidEnvironment"):
                if kwargs.get("verbose"):
                    print("Environment '{}' not found- creating.".format(name),
                          flush=True)
                compute["computeResources"]["subnets"] = net
                compute["computeResources"]["securityGroupIds"] = sg
                compute["computeResources"]["instanceRole"] = self.roles["ecs"]["Arn"].replace("role", "instance-profile")
                compute["serviceRole"] = self.roles["batch"]["Arn"]

                response = self.batch.create_compute_environment(**compute)
                waitUntilDone(name, "CREATING")
                waitUntilDone(name, "UPDATING")
                time.sleep(2)
                compute["computeEnvironmentArn"] = response["computeEnvironmentArn"]

        if kwargs.get("verbose"):
            print("Compute Environment ARN: {}".format(compute["computeEnvironmentArn"]),
                  flush=True)

        template = op.join(op.realpath(clowfile), "templates",
                           "AWS", "jobQueue.json")
        with open(template) as fhandle:
            queue = json.load(fhandle)

        try:
            name = queue["jobQueueName"]
            response = self.batch.describe_job_queues()
            if not len(response["jobQueues"]):
                raise ClientError({"Error": {"Code": "NoSuchEntity"}},
                                  "describe_job_queues")
            else:
                queue_names = [response["jobQueues"][i]["jobQueueName"]
                               for i in range(len(response["jobQueues"]))]
                if name not in queue_names:
                    raise ClientError({"Error": {"Code": "NoSuchEntity"}},
                                      "describe_job_queues")
                queue["jobQueueArn"] = response["jobQueues"][0]["jobQueueArn"]
        except ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchEntity":
                if kwargs.get("verbose"):
                    print("Queue '{}' not found- creating.".format(name),
                          flush=True)
                response = self.batch.create_job_queue(**queue)
                queue["jobQueueArn"] = response["jobQueueArn"]
                time.sleep(2)
        if kwargs.get("verbose"):
            print("Job Queue ARN: {}".format(queue["jobQueueArn"]),
                  flush=True)

        template = op.join(op.realpath(clowfile), "templates",
                           "AWS", "jobDefinition.json")
        with open(template) as fhandle:
            job = json.load(fhandle)

        try:
            name = job["jobDefinitionName"]
            response = self.batch.describe_job_definitions()
            if (not len(response["jobDefinitions"]) or
               response["jobDefinitions"][0]["status"] == "INACTIVE"):
                raise ClientError({"Error": {"Code": "NoSuchEntity"}},
                                  "describe_job_definitions")
            else:
                job["jobDefinitionArn"] = response["jobDefinitions"][0]["jobDefinitionArn"]
        except ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchEntity":
                if kwargs.get("verbose"):
                    print("Job '{}' not found- creating.".format(name),
                          flush=True)
                response = self.batch.register_job_definition(**job)
                job["jobDefinitionArn"] = response["jobDefinitionArn"]

        if kwargs.get("verbose"):
            print("Job Definition ARN: {}".format(job["jobDefinitionArn"]),
                  flush=True)

    def launchJob(self, taskloc):
        # TODO: document
        orides = {"environment":[{"name": "AWS_ACCESS_KEY_ID",
                                  "value": self.access_key},
                                 {"name": "AWS_SECRET_ACCESS_KEY",
                                  "value": self.secret_access}],
                  "command":["task", taskloc, "-V"]}
        # p1, p2 = re.match('.+/.+-([0-9a-zA-Z_]+)/clowdr/task-([A-Za-z0-9]+).json', taskloc).group(1, 2)
        p1, p2 = re.match('.+\/.+-(\w+)\/clowdr\/task-([A-Za-z0-9]+).json',
                          taskloc).group(1, 2)
        response = self.batch.submit_job(jobName="clowdr_{}-{}".format(p1, p2),
                                         jobQueue="clowdr-q",
                                         jobDefinition="clowdr",
                                         containerOverrides=orides)
        jid = response['jobId']
        return jid



#!/usr/bin/env python
#
# This software is distributed with the MIT license:
# https://github.com/gkiar/clowdr/blob/master/LICENSE
#
# clowdr/driver.py
# Created by Greg Kiar on 2018-02-28.
# Email: gkiar@mcin.ca

from argparse import ArgumentParser, RawTextHelpFormatter
import argparse
import os.path as op
import tempfile
import json
import sys
import os

from clowdr.controller import metadata, launcher, rerunner
from clowdr.task import TaskHandler
# from clowdr.server import shareapp, updateIndex
from clowdr.share import consolidate, portal
from clowdr import utils


def local(descriptor, invocation, provdir, backoff_time=36000, **kwargs):
    """cluster
    Launches a pipeline locally through the Clowdr wrappers.

    Parameters
    ----------
    tool : str
        Path to a boutiques descriptor for the tool to be run
    invocation : str
        Path to a boutiques invocation for the tool and parameters to be run
    clowdrloc : str
        Path for storing Clowdr intermediate files and outputs
    dataloc : str
        Path for accessing input data. If local, provide the hostname and
        optionally a path. If on S3, provide an S3 path.
    cluster : str
        Scheduler on the cluster being used. Currently, the only supported mode
        is slurm.
    **kwargs : dict
        Arbitrary keyword arguments. Currently supported arguments:
        - account : str
            Account for the cluster scheduler
        - jobname : str
            Base-name for the jobs as they will appear in the scheduler
        - backoff_time: int
            Time limit for wait times when resubmitting jobs to a scheduler
        - verbose : bool
            Toggle verbose output printing
        - dev : bool
            Toggle dev mode (only runs first execution in the specified set)

        Additionally, transfers all keyword arguments accepted by both of
        "controller.metadata.consolidateTask" and "task.TaskHandler"

    Returns
    -------
    int
        The exit-code returned by the task being executed
    """
    # TODO: scrub inputs
    descriptor = descriptor.name
    tool = utils.truepath(descriptor)
    if kwargs.get("simg"):
        kwargs["simg"] = utils.truepath(kwargs["simg"])

    if kwargs.get("verbose"):
        print("Consolidating metadata...")

    dataloc = kwargs.get("s3") if kwargs.get("s3") else "localhost"
    if kwargs.get("rerun"):
        if not kwargs.get("run_id"):
            raise SystemExit("**Error: Option --rerun requires --run_id")
        tasks = rerunner.getTasks(provdir, kwargs["run_id"], kwargs["rerun"])
        if not len(tasks):
            if kwargs.get("verbose"):
                print("No tasks to run.")
            return 0

    else:
        [tasks, invocs] = metadata.consolidateTask(descriptor, invocation,
                                                   provdir, dataloc, **kwargs)

    taskdir = op.dirname(utils.truepath(tasks[0]))
    try:
        os.mkdir(taskdir)
    except FileExistsError:
        pass
    os.chdir(taskdir)

    with open(tool) as fhandle:
        container = json.load(fhandle).get("container-image")

    if container:
        if kwargs.get("verbose"):
            print("Getting container...")
        outp = utils.getContainer(taskdir, container, **kwargs)

    if kwargs.get("cluster"):
        from slurmpy import Slurm
        jobname = kwargs.get("jobname") if kwargs.get("jobname") else "clowdr"
        slurm_args = {}
        if kwargs.get("slurm_args"):
            for opt in kwargs.get("slurm_args").split(","):
                k, v = opt.split(":")[0], opt.split(":")[1:]
                v = ":".join(v)
                slurm_args[k] = v
        job = Slurm(jobname, slurm_args)

        script = "clowdr task {} -p {} --local"
        if kwargs.get("workdir"):
            script += " -w {}".format(kwargs["workdir"])
        if kwargs.get("volumes"):
            script += " ".join([" -v {}".format(vol)
                                for vol in kwargs.get("volumes")])

    # Groups tasks into collections to be run together (default size = 1)
    gsize = kwargs["groupby"] if kwargs.get("groupby") else 1
    taskgroups = [tasks[i:i+gsize] for i in range(0, len(tasks), gsize)]

    if kwargs.get("dev"):
        taskgroups = [taskgroups[0]]  # Just launch the first in dev mode

    if kwargs.get("verbose"):
        print("Launching tasks...")

    for taskgroup in taskgroups:
        if kwargs.get("verbose"):
            print("... Processing task(s): {}".format(", ".join(taskgroup)))

        if kwargs.get("cluster"):
            tmptaskgroup = " ".join(taskgroup)
            func = job.run
            args = [script.format(tmptaskgroup, taskdir)]
            # Submit. If submission fails, retry with fibonnaci back-off
            utils.backoff(func, args, backoff_time=backoff_time, **kwargs)
        else:
            runtask(taskgroup, provdir=taskdir, local=True, **kwargs)

    if kwargs.get("verbose"):
        print(taskdir)
    return taskdir


def cloud(descriptor, invocation, provdir, s3, cloud, credentials, **kwargs):
    """cloud
    Launches a pipeline locally at scale through Clowdr.

    Parameters
    ----------
    descriptor : str
        Path to a boutiques descriptor for the tool to be run
    invocation : str
        Path to a boutiques invocation for the tool and parameters to be run
    provdir : str
        Path on S3 for storing Clowdr intermediate files and outputs
    s3 : str
        Path on S3 for accessing input data
    cloud : str
        Which endpoint to use for deployment
    credentials : str
        Credentials for Amazon with access to dataloc, clowdrloc, and Batch
    **kwargs : dict
        Arbitrary keyword arguments (i.e. {'verbose': True})

    Returns
    -------
    int
        The exit-code returned by the task being executed
    """
    # TODO: scrub inputs better
    descriptor = descriptor.name
    provdir = provdir.strip('/')

    # Create temp dir for clowdrloc
    tmploc = utils.truepath(tempfile.mkdtemp())

    [tasks, invocs] = metadata.consolidateTask(descriptor, invocation, tmploc,
                                               s3, **kwargs)
    metadata.prepareForRemote(tasks, tmploc, provdir)
    resource = launcher.configureResource(cloud, credentials, **kwargs)

    tasks_remote = [task for task in utils.post(tmploc, provdir)
                    if "task-" in task]

    if kwargs.get("dev"):
        tasks_remote = [tasks_remote[0]]  # Just launch the first in dev mode

    jids = []
    for task in tasks_remote:
        jids += [resource.launchJob(task)]

    taskdir = op.dirname(utils.truepath(tasks_remote[0]))
    print(taskdir)
    return taskdir, jids


def runtask(tasklist, **kwargs):
    for task in tasklist:
        handler = TaskHandler(task, **kwargs)


def share(provdir, **kwargs):
    """share
    Launches a simple web server which showcases all runs at the clowdrloc.

    Parameters
    ----------
    provdir : str
        Path with Clowdr metdata files (returned from "local" and "deploy")
    **kwargs : dict
        Arbitrary keyword arguments (i.e. {'verbose': True})

    Returns
    -------
    None
    """
    if provdir.startswith("s3://"):
        # Create temp dir for clowdrloc
        tmploc = utils.truepath(tempfile.mkdtemp())
        utils.get(provdir, tmploc, **kwargs)
        tmpdir = op.join(tmploc, utils.splitS3Path(provdir)[1])
        provdir = tmpdir
        if kwargs.get("verbose"):
            print("Local cache of directory: {}".format(provdir))

    if op.isfile(provdir):
        if kwargs.get("verbose"):
            print("Summary file provided - no need to generate.")
        summary = provdir
        with open(summary) as fhandle:
            experiment_dict = json.load(fhandle)
    else:
        summary = op.join(provdir, 'clowdr-summary.json')
        experiment_dict = consolidate.summary(provdir, summary)

    customDash = portal.CreatePortal(experiment_dict)
    app = customDash.launch()

    host = kwargs["host"] if kwargs.get("host") else "0.0.0.0"
    app.run_server(host=host, debug=kwargs.get("debug"))


def makeparser():
    """makeparser
    Command-line API wrapper for Clowdr as a CLI, not Python API.
    For information about the command-line wrapper and arguments it accepts,
    please try running "clowdr --help".

    Parameters
    ----------
    args: list
        List of all command-line arguments being passed.

    Returns
    -------
    int
        The exit-code returned by the driver.
    """

    # Create an outer argparser which can wrap subparsers for each function.
    desc = """
Scalable deployment and provenance-rich wrapper for Boutiques tools locally,
on clusters, and in the cloud. For more information, go to our website:

     https://github.com/clowdr/clowdr.
"""
    parser = ArgumentParser("clowdr", description=desc,
                            formatter_class=RawTextHelpFormatter)

    htext = """Clowdr has several distinct modes of operation:
  - local:  This mode allows you to develop your Clowdr execution, deploy
            analyses on your local system, and deploy them on clusters.
  - cloud:  This mode allows you to deploy your Clowdr exectuion on a cloud
            resource. Currently, this only supports Amazon Web Services.
  - share:  This mode launches a lightweight webserver for you to explore your
            executions, monitor job progress, and share your results.
  - task:   This mode is generally only for super-users. It is used by Clowdr
            to launch your tasks and record provenance information from them
            without you needing to call this option yourself. It can be useful
            when debugging or re-running failed executions.
"""
    subparsers = parser.add_subparsers(dest="mode",
                                       help=htext)

    # Create the subparser for local/cluster execution.
    desc = ("Manages local and cluster deployment. Ideal for development, "
            "testing, executing on local resources, or deployment on a "
            "computing cluster environment.")
    parser_loc = subparsers.add_parser("local", description=desc)
    parser_loc.add_argument("descriptor", type=argparse.FileType('r'),
                            help="Local path to Boutiques descriptor for the "
                                 "tool you wish to run. To learn about "
                                 "descriptors and Boutiques, go to: "
                                 "https://boutiques.github.io.")
    parser_loc.add_argument("invocation",
                            help="Local path to Boutiques invocation (or "
                                 "directory containing multiple invocations) "
                                 "for the analysis you wish to run. To learn "
                                 "about invocations and Boutiques, go to: "
                                 "https://boutiques.github.io.")
    parser_loc.add_argument("provdir",
                            help="Local directory for Clowdr provenance records"
                                 " and other captured metadata to be stored. "
                                 "This directory needs to exist prior to "
                                 "running Clowdr.")

    parser_loc.add_argument("--verbose", "-V", action="store_true",
                            help="Toggles verbose output statements.")
    parser_loc.add_argument("--dev", "-d", action="store_true",
                            help="Launches only the first created task. This "
                                 "is intended for development purposes.")
    parser_loc.add_argument("--workdir", "-w", action="store",
                            help="Specifies the working directory to be used "
                                 "by the tasks created.")
    parser_loc.add_argument("--volumes", "-v", action="append",
                            help="Specifies any volumes to be mounted to the "
                                 "container. This is usually related to the "
                                 "path of any data files as specified in your "
                                 "invocation(s).")
    parser_loc.add_argument("--groupby", "-g", type=int,
                            help="If you wish to run tasks in batches, specify "
                                 "the number of tasks to group here. For "
                                 "imperfect multiples, the last group will be "
                                 "the remainder.")
    parser_loc.add_argument("--cluster", "-c", choices=["slurm"],
                            help="If you wish to submit your local tasks to a "
                                 "scheduler, you must specify it here. "
                                 "Currently this only supports SLURM clusters.")
    parser_loc.add_argument("--clusterargs", "-a", action="store",
                            help="This allows users to supply arguments to the "
                                 "cluster, such as specifying RAM or requesting"
                                 " a certain amount of time on CPU. These are "
                                 "provided in the form of key:value pairs, and "
                                 "separated by commas. For example: "
                                 "--clusterargs time:4:00,mem:2048,account:ABC")
    parser_loc.add_argument("--jobname", "-n", action="store",
                            help="If running on a cluster, and you wish to "
                                 "specify a unique identifier to appear in the"
                                 "submitted tasks, you can specify it with "
                                 "this flag.")
    parser_loc.add_argument("--simg", "-s", action="store",
                            help="If the Boutiques descriptor summarizes a "
                                 "tool wrapped in Singularity, and the image "
                                 "has already been downloaded, this option "
                                 "allows you to specify that image file.")
    parser_loc.add_argument("--user", "-u", action="store_true",
                            help="If the Boutiques descriptor summarizes a "
                                 "tool wrapped in Docker, toggles propagating "
                                 "the current user within the container.")
    parser_loc.add_argument("--rerun", "-R",
                            choices=["all", "failed", "incomplete"],
                            help="Allows user to re-run jobs in a previous "
                                 "execution that either failed or didn't "
                                 "finish, etc. This requires the --run_id "
                                 "argument to also be supplied. Three choices "
                                 "are: 'all' to re-run all tasks, 'failed' to "
                                 "re-run tasks which finished with a non-zero "
                                 "exit-code, 'incomplete' to re-run tasks "
                                 "which have not yet indicated job completion."
                                 " While the descriptor and invocations will be"
                                 " adopted from the previous executions, other "
                                 "options such as clusterargs or volume can "
                                 "be set to different values, if they were the "
                                 "source or errors. Pairing the incomplete mode"
                                 " with the --dev flag allows you to walk "
                                 "through your dataset one group at a time.")
    parser_loc.add_argument("--run_id", action="store",
                            help="Pairs with --rerun. This ID is the directory"
                                 " within the supplied provdir which contains "
                                 "execution you wish to relaunch. These IDs/"
                                 "directories are in the form: year-month-day_"
                                 "hour-minute-second-8digitID.")
    parser_loc.add_argument("--s3", action="store",
                            help="Amazon S3 bucket and path for remote data. "
                                 "Accepted in the format: s3://{bucket}/{path}")
    parser_loc.add_argument("--bids", "-b", action="store_true",
                            help="Indicates that the tool being launched is a "
                                 "BIDS app. BIDS is a data organization format"
                                 " in neuroimaging. For more information about"
                                 " this, go to https://bids.neuroimaging.io.")

    parser_loc.set_defaults(func=local)

    # Create the subparser for cloud execution.
    desc = ("Manages cloud deployment. Ideal for running jobs at scale on data "
            "stored in Amazon Web Services S3 buckets (or similar object "
            "store).")
    parser_cld = subparsers.add_parser("cloud", description=desc)
    parser_cld.add_argument("descriptor", type=argparse.FileType('r'),
                            help="Local path to Boutiques descriptor for the "
                                 "tool you wish to run. To learn about "
                                 "descriptors and Boutiques, go to: "
                                 "https://boutiques.github.io.")
    parser_cld.add_argument("invocation",
                            help="Local path to Boutiques invocation (or "
                                 "directory containing multiple invocations) "
                                 "for the analysis you wish to run. To learn "
                                 "about invocations and Boutiques, go to: "
                                 "https://boutiques.github.io.")
    parser_cld.add_argument("provdir",
                            help="Local directory for Clowdr provenance records"
                                 " and other captured metadata to be stored. "
                                 "This directory needs to exist prior to "
                                 "running Clowdr.")
    parser_cld.add_argument("s3",
                            help="Amazon S3 bucket and path for remote data. "
                                 "Accepted in the format: s3://{bucket}/{path}")
    parser_cld.add_argument("cloud", choices=["aws"],
                            help="Specifies which cloud endpoint you'd like to"
                                 " use. Currently, only AWS is supported.")
    parser_cld.add_argument("credentials",
                            help="Your credentials file for the resource.")

    parser_cld.add_argument("--verbose", "-V", action="store_true",
                            help="Toggles verbose output statements.")
    parser_cld.add_argument("--dev", "-d", action="store_true",
                            help="Launches only the first created task. This "
                                 "is intended for development purposes.")
    parser_cld.add_argument("--region", "-r", action="store",
                            help="The Amazon region to use for processing.")
    parser_cld.add_argument("--bids", "-b", action="store_true",
                            help="Indicates that the tool being launched is a "
                                 "BIDS app. BIDS is a data organization format"
                                 " in neuroimaging. For more information about"
                                 " this, go to https://bids.neuroimaging.io.")

    parser_cld.set_defaults(func=cloud)

    # Create the subparser for sharing outputs
    desc = ("Launches light-weight web service for exploring, managing, and "
            "sharing the outputs and provenance recorded from Clowdr "
            "executed workflows.")
    parser_shr = subparsers.add_parser("share")
    parser_shr.add_argument("provdir",
                            help="Local or S3 directory where Clowdr provenance"
                                 "records and metadata are stored. This path "
                                 "was returned by running either clowdr cloud "
                                 "or clowdr local. This can also be a clowdr-"
                                 "generated summary file.")
    parser_shr.add_argument("--debug", "-d", action="store_true",
                            help="Toggles server messages and logging. This "
                                 "is intended for development purposes.")
    parser_shr.add_argument("--verbose", "-V", action="store_true",
                            help="Toggles verbose output statements.")

    parser_shr.set_defaults(func=share)

    # Create the subparser for launching tasks
    desc = ("Launches a list of tasks with provenance recording. This method "
            "is what specifically wraps tool execution, is called by other "
            "Clowdr modes, and can be used to re-execute or debug tasks.")
    parser_task = subparsers.add_parser("task")
    parser_task.add_argument("tasklist", nargs="+",
                             help="One or more Clowdr-created task.json files "
                                  "summarizing the jobs to be run. These task "
                                  "files are created by one of clowdr cloud or"
                                  " clowdr local.")

    parser_task.add_argument("--verbose", "-V", action="store_true",
                             help="Toggles verbose output statements.")
    parser_task.add_argument("--provdir", "-p", action="store",
                             help="Local or directory where Clowdr provenance "
                                  "records and metadata will be stored. This "
                                  "is optional here because it will be stored "
                                  "by default in a temporary location and "
                                  "moved, unless this is specified.")
    parser_task.add_argument("--local", "-l", action="store_true",
                             help="Flag indicator to identify whether the task"
                                  " is being launched on a cloud or local "
                                  "resource. This is important to ensure data "
                                  "is transferred off clouds before shut down.")
    parser_task.add_argument("--workdir", "-w", action="store",
                             help="Specifies the working directory to be used "
                                  "by the tasks created.")
    parser_task.add_argument("--volumes", "-v", action="append",
                             help="Specifies any volumes to be mounted to the "
                                  "container. This is usually related to the "
                                  "path of any data files as specified in your "
                                  "invocation(s).")

    parser_task.set_defaults(func=runtask)
    return parser


def main(args=None):
    parser = makeparser()

    # Parse arguments
    inps = parser.parse_args(args) if args is not None else parser.parse_args()

    # If no args are provided, print help
    if len(sys.argv) < 2 and args is None:
        parser.print_help()
        sys.exit()
    else:
        inps.func(**vars(inps))
        return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# This software is distributed with the MIT license:
# https://github.com/gkiar/clowdr/blob/master/LICENSE
#
# clowdr/controller/launcher.py
# Created by Greg Kiar on 2018-02-28.
# Email: gkiar@mcin.ca


# from clowdr.endpoint import aws, kubernetes, cbrain

from clowdr import utils


def configureResource(endpoint, auth, **kwargs):
    # TODO: document
    if endpoint == "aws":
        from clowdr.endpoint.AWS import AWS
        resource = AWS(auth)
        resource.setCredentials(**kwargs)
        resource.startSession()
        resource.configureIAM(**kwargs)
        resource.configureBatch(**kwargs)
        return resource

    elif endpoint == "kubernetes":
        print("Kubernetes endpoint not yet supported - coming soon!")

    elif endpoint == "cbrain":
        print("CBRAIN endpoint not yet supported - coming soon!")

    else:
        print("Endpoint not currently supported. Try: aws".format(endpoint))




#!/usr/bin/env python

from dash.dependencies import Input, Output, State
import dash_core_components as dcc
import dash_html_components as html
import dash_table_experiments as dt
import dash

import plotly.figure_factory as ff
import plotly.graph_objs as go
import plotly

from datetime import datetime, timedelta
import pandas as pd
import numpy as np
import json

from clowdr.share import customDash as cd


class CreatePortal():
    def __init__(self, experiment_dict):
        # Initialize Dash app with custom wrapper that lets us set page title
        self.app = cd.CustomDash()
        # Load and groom dataset
        # -----> /start data grooming

        self.main_colour = '#0074D9'
        self.accent_colour = '#FF851B'

        # Intitialize two new data tables for visualization, resulting in:
        #   - (Full) experiment table, to be used in visualizations
        #   - Stats table, to be used for filtering based on summary info
        #   - Invocation table, to be used for filtering based on invoc. info
        self.stat_dict = []
        self.invo_dict = []
        new_experiment_dict = []
        for exp in experiment_dict:
            # Coerce task ID into integer value
            exp['Task ID'] = int(exp['Task ID'])
            if exp['Time: Start'] is not None:
                time_s = datetime.strptime(exp['Time: Start'], "%Y-%m-%d %H:%M:%S")
                time_e = time_s + timedelta(0, exp['Time: Total (s)'])
            else:
                time_e = None
            exp['Time: End'] = str(time_e)

            # Add relevant info to stats dict
            self.stat_dict += [{'Task': exp['Task ID'],
                                'Tool': exp['Tool Name'],
                                'Duration (s)': exp['Time: Total (s)'],
                                'Max RAM (MB)': exp['RAM: Max (MB)'],
                                'Max CPU (%)': exp['CPU: Max (%)'],
                                'Exit Code': exp['Exit Code']}]

            # Add relevant info to invocation dict
            tmp_dict = {'Task': exp['Task ID']}
            tmp_dict.update({key.replace('Param: ', ''): exp[key]
                             for key in exp.keys()
                             if key.startswith('Param: ')})
            tmp_dict = {k: str(tmp_dict[k])
                        if type(tmp_dict[k]) == bool
                        else tmp_dict[k]
                        for k in tmp_dict.keys()}
            self.invo_dict += [tmp_dict]

            # Save the minor updates back to the dictionary
            new_experiment_dict += [exp]

        # Get rid of un-updated version of the dictionary
        self.experiment_dict = new_experiment_dict

        # Create look-up-table for mapping tasks to rows in the tables
        self.idx_to_task_lut = {exp['Task ID']: idx
                                for idx, exp in enumerate(experiment_dict)}
        self.task_to_idx_lut = {idx: exp['Task ID']
                                for idx, exp in enumerate(experiment_dict)}
        # <------ /stop data grooming

    def launch(self):
        self.create_dashboard()
        self.create_callbacks()
        return self.app

    def create_dashboard(self):
        # Create page objects and their content
        # ------> /start page creation
        config = {'modeBarButtonsToRemove': ['sendDataToCloud']}
        self.app.layout = html.Div([
            # Page title/header
            html.H4('Clowdr Experiment Explorer'),

            # Tabs & tables
            dcc.Tabs(id='tabs', value='stats-tab',
                     children=self.create_tabs_children('stats-tab',
                                                        [0],
                                                        self.stat_dict)),

            # Gantt container (to be updated by callbacks)
            # dcc.Graph(id='gantt-clowdrexp',
            #           figure=self.create_gantt(self.stat_dict, [0]),
            #           config=config),
            # TODO create gantt function; new callback to update gantt area

            # Graph container (to be populated by callbacks)
            dcc.Graph(id='graph-clowdrexp',
                      figure=self.create_figure(self.stat_dict, [0]),
                      config=config)
            ], className="container")
        # <------ /stop page creation

    def create_callbacks(self):
        # Create callbacks for page elements
        # ------> /start callback management

        # Callback: update table based on tab & preserve selected rows
        @self.app.callback(
            Output('tabs', 'children'),
            [Input('tabs', 'value')],
            [State('table-clowdrexp', 'selected_row_indices'),
             State('table-clowdrexp', 'rows')])
        def redraw_table(tab, selected_indices, rows):
            return self.create_tabs_children(tab, selected_indices, rows)

        # Callback: update selected rows based on available rows or click event
        @self.app.callback(
            Output('table-clowdrexp', 'selected_row_indices'),
            [Input('graph-clowdrexp', 'clickData')],
            [State('table-clowdrexp', 'selected_row_indices'),
             State('table-clowdrexp', 'rows')])
        def update_selected_rows(clickData, selected_indices, rows):
            # global_indices = self.get_global_index(rows, selected_indices)
            if clickData:
                point = clickData['points'][0]
                if point['customdata'] == 'task':
                    curve = point['curveNumber'] - (len(selected_indices) * 2)
                    if curve in selected_indices:
                        selected_indices.remove(curve)
                    else:
                        selected_indices.append(curve)
            return selected_indices

        # Callback: update figure based on selected/present data
        @self.app.callback(
            Output('graph-clowdrexp', 'figure'),
            [Input('table-clowdrexp', 'selected_row_indices')],
            [State('table-clowdrexp', 'rows')])
        def update_figure(selected_indices, rows):
            # Create new figure based on selected rows
            global_indices = self.get_global_index(rows, selected_indices)
            figure = self.create_figure(rows, global_indices)
            return figure
    # <------ /stop callback management

    # Create helper functions to be used by callbacks & page creation
    # ------> /start helper function creation
    # Creates the data table given the dictionary and selected rows
    def create_datatable(self, data_dict, selected_indices):

        return dt.DataTable(
                    rows=data_dict,  # Table data
                    columns=list(data_dict[0].keys()),  # Column names
                    row_selectable=True,  # Able to select independent rows
                    filterable=True,  # Able to filter columns by value
                    # debounced=True,  # Delay after filtering before callback
                    sortable=True,  # Able to sort by column values
                    editable=False,  # Not able to edit values
                    resizable=True,  # Able to resize columns to fit data
                    max_rows_in_viewport=5,  # Scroll if more than 5 rows
                    selected_row_indices=selected_indices,  # selected rows
                    id='table-clowdrexp')  # Use the same ID for callbacks

    # Redraws the two-tab layout and table, preserving selected data.
    #  - Tab 1: Stats table
    #  - Tab 2: Invocation table
    def create_tabs_children(self, active_tab, selected_indices, rows):
        # If the table has reduced rows, we need to map the rows to tasks
        global_indices = self.get_global_index(rows, selected_indices)

        # For each tab, recreate the table using the correct indices and data
        if active_tab == 'stats-tab':
            stat_child = [self.create_datatable(self.stat_dict, global_indices)]
            invo_child = []
        elif active_tab == 'invo-tab':
            invo_child = [self.create_datatable(self.invo_dict, global_indices)]
            stat_child = []

        # Put the table and empty table into the layout
        tabs = [
                # Stat table creation
                dcc.Tab(label="Statistics",
                        value='stats-tab',
                        children=stat_child),
                # Invocation table creation
                dcc.Tab(label="Invocations",
                        value='invo-tab',
                        children=invo_child)]
        return tabs

    def create_figure(self, rows, global_indices):
        plotting_data = [self.experiment_dict[idx] for idx in global_indices]
        row_ids = [r['Task'] for r in rows]

        # Initialize plotting space
        fig = plotly.tools.make_subplots(rows=3, cols=1,
                                         subplot_titles=('RAM Usage',
                                                         'CPU Usage',
                                                         'Tasks'))

        data = []
        id_list = []
        opacity = 2/(len(plotting_data)+1)
        for i, exp in enumerate(plotting_data):
            if exp['Exit Code'] == 'Incomplete':
                continue
            id_list += [exp['Task ID']]
            #1 fig = self.append_trace(fig, exp, i)
            data += self.append_trace(exp, i, opacity)

        # Create the generic gantt chart
        ganttdat = []
        for i, exp in enumerate(self.experiment_dict):
            if exp['Exit Code'] == 'Incomplete':
                continue
            tmpfig = ff.create_gantt([{'Task': exp['Task ID'],
                                       'Start': exp['Time: Start'],
                                       'Finish': exp['Time: End']}])
            t3 = tmpfig['data'][0]
            t3['y'] = (i, i)
            t3['yaxis'] = 'y3'
            t3['customdata'] = ['task'] * 2
            t3['hoverinfo'] = 'name'
            t3['xaxis'] = 'x2'
            t3['mode'] = 'lines'
            if i in global_indices:
                tcolour = self.accent_colour
            else:
                tcolour = self.main_colour

            if exp['Task ID'] in row_ids:
                twidth = 5
            else:
                twidth = 2
            t3['line'] = {'color': tcolour, 'width': twidth}
            t3['name'] = 'Task {}'.format(exp['Task ID'])
            ganttdat += [t3]

        data += ganttdat
        layout = {
            'showlegend': False,
            'height': 600,
            'xaxis': {
                'title': 'Time (s)',
                'domain': [0, 1]
            },
            'xaxis2': {
                'title': 'Datetime',
                'anchor': 'y3',
                'domain': [0, 1]
            },
            'yaxis': {
                'title': 'RAM (MB)',
                'rangemode': 'tozero',
                'domain': [0.35, 0.7]
            },
            'yaxis2': {
                'title': 'CPU (%)',
                'autorange': 'reversed',
                'rangemode': 'tozero',
                'side': 'left',
                'domain': [0.0, 0.35]
            },
            'yaxis3': {
                'title': 'Task',
                'autorange': 'reversed',
                'zeroline': False,
                'domain': [0.85, 1],
            },
            'hovermode': 'closest'
        }

        fig = go.Figure(data, layout)
        return fig

    # Utility for adding a trace back to the graph
    def append_trace(self, data_row, idx, opacity=0.6):
        # RAM Time series plot
        t1 = go.Scattergl(
            x=data_row['Time: Series (s)'],
            y=data_row['RAM: Series (MB)'],
            mode='lines+markers',
            line={'color': self.main_colour},
            opacity=opacity,
            fill='tozeroy',
            xaxis='x',
            yaxis='y',
            customdata=['ram'] * len(data_row['Time: Series (s)']),
            name='Task {}'.format(data_row['Task ID'])
        )

        # CPU Time series plot
        t2 = go.Scattergl(
            x=data_row['Time: Series (s)'],
            y=data_row['CPU: Series (%)'],
            mode='lines+markers',
            line={'color': self.accent_colour},
            opacity=opacity,
            fill='tozeroy',
            xaxis='x',
            yaxis='y2',
            customdata=['cpu'] * len(data_row['Time: Series (s)']),
            name='Task {}'.format(data_row['Task ID'])
        )

        data = [t1, t2]
        return data

    # Utility for getting the global index from index on subset
    def get_global_index(self, rows, selected_indices):

        return [self.task_to_idx_lut[row['Task']]
                for idx, row in enumerate(rows)
                if idx in selected_indices]

    # Utility to get all traces corresponding to a particular task ID
    def get_trace_locations(self, traces, task_id):

        return [idx for idx, tr in enumerate(traces)
                if tr['name'] == 'Task {}'.format(task_id)]
    # <------ /stop helper function creation



#!/usr/bin/env python
#
# This software is distributed with the MIT license:
# https://github.com/gkiar/clowdr/blob/master/LICENSE
#
# clowdr/controller/metadata.py
# Created by Greg Kiar on 2018-02-28.
# Email: gkiar@mcin.ca

from copy import deepcopy
import os.path as op
import datetime
import time
import string
import json
import sys
import os

from clowdr import utils


def consolidateTask(tool, invocation, clowdrloc, dataloc, **kwargs):
    """consolidate
    Creates Clowdr task JSON files which summarize all associated metadata

    Parameters
    ----------
    tool : str
        Path to a boutiques descriptor for the tool to be run
    invocation : str
        Path to a boutiques invocation for the tool and parameters to be run
    clowdrloc : str
        Path for storing Clowdr intermediate files and outputs
    dataloc : str
        Path for accessing input data
    **kwargs : dict
        Arbitrary keyword arguments (i.e. {'verbose': True})

    Returns
    -------
    tuple: (list, list)
        The task dictionary JSONs, and associated Boutiques invocation files.
    """

    ts = time.time()
    dt = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d_%H-%M-%S')
    randx = utils.randstring(8)
    modif = "{}-{}".format(dt, randx)

    # Scrub inputs
    tool = utils.truepath(tool)
    invocation = utils.truepath(invocation)
    clowdrloc = utils.truepath(clowdrloc)
    dataloc = utils.truepath(dataloc)

    # Initialize task dictionary
    taskdict = {}
    with open(tool) as fhandle:
        toolname = json.load(fhandle)["name"].replace(' ', '-')
    taskloc = op.join(clowdrloc, modif, 'clowdr')
    os.makedirs(taskloc)

    taskdict["taskloc"] = op.join(clowdrloc, modif, toolname)
    taskdict["dataloc"] = [dataloc]
    taskdict["invocation"] = utils.get(invocation, taskloc)[0]
    taskdict["tool"] = utils.get(tool, taskloc)[0]

    # Case 1: User supplies directory of invocations
    if op.isdir(invocation):
        tempinvocations = os.listdir(invocation)
        taskdicts = []
        invocations = []
        for invoc in tempinvocations:
            tempdict = deepcopy(taskdict)
            tempinvo = utils.get(op.join(invocation, invoc), taskloc)
            tempdict["invocation"] = utils.truepath(tempinvo[0])
            invocations += tempinvo
            taskdicts += [tempdict]

    # Case 2: User supplies a single invocation
    else:
        # Case 2a: User is running a BIDS app
        if kwargs.get("bids"):
            taskdicts, invocations = bidsTasks(taskloc, taskdict)

        # Case 2b: User is quite simply just launching a single invocation
        else:
            taskdicts = [taskdict]
            invocations = [taskdict["invocation"]]

    # Store task definition files to disk
    taskdictnames = []
    for idx, taskdict in enumerate(taskdicts):
        taskfname = op.join(taskloc, "task-{}.json".format(idx))
        taskdictnames += [taskfname]
        with open(taskfname, 'w') as fhandle:
            fhandle.write(json.dumps(taskdict))

    return (taskdictnames, invocations)


def bidsTasks(clowdrloc, taskdict):
    """bidsTask
    Scans through BIDS app fields for creating more tasks than specified.

    Parameters
    ----------
    clowdrloc : str
        Path for storing Clowdr intermediate files and outputs
    taskdict : str
        Dictionary of the tasks (pre-BIDS-ification)

    Returns
    -------
    tuple: (list, list)
        The task dictionary JSONs, and associated Boutiques invocation files.
    """

    dataloc = taskdict["dataloc"][0]
    invocation = taskdict["invocation"]

    invo = json.load(open(invocation))
    participants = invo.get("participant_label")
    sessions = invo.get("session_label")

    # Case 1: User is running BIDS group-level analysis
    if invo.get("analysis_level") == "group":
        return ([taskdict], [invocation])

    # Case 2: User is running BIDS participant- or session-level analysis
    #       ... and specified neither participant(s) nor session(s)
    elif not participants and not sessions:
        return ([taskdict], [invocation])

    # Case 3: User is running BIDS participant- or session-level analysis
    #       ... and specified participant(s) but not session(s)
    elif participants and not sessions:
        taskdicts = []
        invos = []
        for part in participants:
            partstr = "sub-{}".format(part)

            tempdict = deepcopy(taskdict)
            tempdict["dataloc"] = [op.join(dataloc, partstr)]

            invo["participant_label"] = [part]

            invofname = op.join(clowdrloc, "invocation_sub-{}.json".format(part))
            with open(invofname, 'w') as fhandle:
                fhandle.write(json.dumps(invo))

            tempdict["invocation"] = invofname
            invos += [invofname]
            taskdicts += [tempdict]
        return (taskdicts, invos)

    # Case 4: User is running BIDS participant- or session-level analysis
    #       ... and specified participants(s) and session(s)
    elif participants and sessions:
        taskdicts = []
        invos = []
        for part in participants:
            partstr = "sub-{}".format(part)

            for sesh in sessions:
                seshstr = "ses-{}".format(sesh)

                tempdict = deepcopy(taskdict)
                tempdict["dataloc"] = [op.join(dataloc, partstr, seshstr)]

                invo["participant_label"] = [part]
                invo["session_label"] = [sesh]

                invofname = op.join(clowdrloc, "invocation_sub-{}_ses-{}.json".format(part, sesh))
                with open(invofname, 'w') as fhandle:
                    fhandle.write(json.dumps(invo))

                tempdict["invocation"] = invofname
                invos += [invofname]
                taskdicts += [tempdict]
        return (taskdicts, invos)

    # Case 5: User is running BIDS participant- or session-level analysis
    #       ... and specified sessions(s) but not participant(s)
    else:
        print("Error: Invalid BIDS mode not supported....")
        print("       Must specify participant label if specifying sessions!")
        sys.exit(-1)


def prepareForRemote(tasks, tmploc, clowdrloc):
    """prepare
    Scans through BIDS app fields for creating more tasks than specified.

    Parameters
    ----------
    clowdrloc : str
        Path for storing Clowdr intermediate files and outputs
    taskdict : str
        Dictionary of the tasks (pre-BIDS-ification)

    Returns
    -------
    tuple: (list, list)
        The task dictionary JSONs, and associated Boutiques invocation files.
    """

    # Modify tasks
    for task in tasks:
        with open(task) as fhandle:
            task_dict = json.load(fhandle)

        task_dict["invocation"] = op.join(clowdrloc,
                                          op.relpath(task_dict["invocation"],
                                                     tmploc))
        task_dict["taskloc"] = op.join(clowdrloc,
                                       op.relpath(task_dict["taskloc"],
                                                  tmploc))
        task_dict["tool"] = op.join(clowdrloc, op.relpath(task_dict["tool"],
                                                          tmploc))

        with open(task, 'w') as fhandle:
            fhandle.write(json.dumps(task_dict))

    return 0




#/usr/bin/env python

import dash

class CustomDash(dash.Dash):
    def interpolate_index(self, **kwargs):
        # Inspect the arguments by printing them
        return '''
<!DOCTYPE html>
<html>
    <head>
    <title>Clowdr Explorer</title>
    <link rel="stylesheet" href="https://unpkg.com/react-select@1.0.0-rc.3/dist/react-select.min.css">
    <link rel="stylesheet" href="https://unpkg.com/react-virtualized@9.9.0/styles.css">
    <link rel="stylesheet" href="https://unpkg.com/react-virtualized-select@3.1.0/styles.css">
    <link rel="stylesheet" href="https://unpkg.com/rc-slider@6.1.2/assets/index.css">
    <link rel="stylesheet" href="https://unpkg.com/dash-core-components@0.27.1/dash_core_components/react-dates@12.3.0.css">
    <link rel="stylesheet" href="https://unpkg.com/dash-table-experiments@0.6.0/dash_table_experiments/dash_table_experiments.css">
    <link rel="stylesheet" href="https://codepen.io/chriddyp/pen/bWLwgP.css">
    <link rel="stylesheet" href="/assets/banner.css?mod=1534778913.0">
    </head>
    <body>
        {app_entry}
        {config}
        {scripts}
    <footer>
    <script src="https://unpkg.com/react@15.4.2/dist/react.min.js"></script>
    <script src="https://unpkg.com/react-dom@15.4.2/dist/react-dom.min.js"></script>
    <script src="https://unpkg.com/dash-html-components@0.11.0/dash_html_components/bundle.js"></script>
    <script src="https://cdn.plot.ly/plotly-1.39.1.min.js"></script>
    <script src="https://unpkg.com/dash-core-components@0.27.1/dash_core_components/bundle.js"></script>
    <script src="https://unpkg.com/dash-table-experiments@0.6.0/dash_table_experiments/bundle.js"></script>
    <script src="https://unpkg.com/dash-renderer@0.13.0/dash_renderer/bundle.js"></script>
    </footer>
    </body>
</html>
        '''.format(
            app_entry=kwargs['app_entry'],
            config=kwargs['config'],
            scripts=kwargs['scripts'])




#!/usr/bin/env python

import os.path as op
import pandas as pd
import numpy as np
import json
import os
import re


def summary(indir, outfile):
    # Get list of tasks
    tasks = [op.join(indir, f)
             for f in os.listdir(indir)
             if re.match(r'task-[0-9]+.json', f)]
    experiment = []

    # For each task...
    for idx, task_file in enumerate(tasks):
        task_id = re.findall(r'.+task-([0-9]+).json', task_file)[0]
        task_dict = {}

        # Load task file...
        with open(task_file) as task_fhandle:
            # ... and extract the descriptor and invocation
            tmp_task = json.load(task_fhandle)
            descriptor_file = op.join(indir, op.basename(tmp_task['tool']))
            invocat_file = op.join(indir, op.basename(tmp_task['invocation']))

        # Load descriptor file...
        with open(descriptor_file) as descriptor_fhandle:
            # ... and extract the tool name and list of inputs
            tmp_desc = json.load(descriptor_fhandle)
            tmp_name = tmp_desc['name']
            tmp_inps = {inp['id']: inp['name'] for inp in tmp_desc['inputs']}

        # Load the invocation file...
        with open(invocat_file) as invocation_fhandle:
            # ... and extract the invocation parameters used
            tmp_invo = json.load(invocation_fhandle)
            tmp_invo_dict = {}
            for inp in tmp_inps:
                key = 'Param: {}'.format(tmp_inps[inp])
                if tmp_invo.get(inp):
                    tmp_invo_dict[key] = tmp_invo[inp]

        # Load the summary file...
        summary_file = op.join(indir, 'task-' + task_id + '-summary.json')
        if op.isfile(summary_file):
            with open(summary_file) as summary_fhandle:
                # ... and extract the exit code and duration
                tmp_summ = json.load(summary_fhandle)
                tmp_time = tmp_summ['launchtime']
                tmp_ecod = tmp_summ['exitcode']
                tmp_durr = tmp_summ['duration']
        else:
            tmp_summ = None
            tmp_time = None
            tmp_ecod = "Incomplete"
            tmp_durr = None

        # Load stdout...
        stdout_file = op.join(indir, 'task-' + task_id + '-stdout.txt')
        if op.isfile(stdout_file):
            with open(stdout_file) as stdout_fhandle:
                # ... and save it
                tmp_sout = stdout_fhandle.read()
        else:
            tmp_sout = None

        # Load stderr...
        stderr_file = op.join(indir, 'task-' + task_id + '-stderr.txt')
        if op.isfile(stderr_file):
            with open(stderr_file) as stderr_fhandle:
                # ... and save it
                tmp_serr = stderr_fhandle.read()
        else:
            tmp_serr = None

        # Load the usage file...
        usage_file = op.join(indir, 'task-' + task_id + '-usage.csv')
        if op.isfile(usage_file):
            tmp_df = pd.read_csv(usage_file)
            # ... and extract the RAM time series and summary of it
            tmp_tim = [w for v in tmp_df[['time']].values.tolist() for w in v]
            tmp_ram = [w for v in tmp_df[['ram']].values.tolist() for w in v]
            tmp_cpu = [w for v in tmp_df[['cpu']].values.tolist() for w in v]
            tmp_max_ram = np.max(tmp_ram)
            tmp_max_cpu = np.max(tmp_cpu)
        else:
            tmp_tim = None
            tmp_ram = None
            tmp_cpu = None
            tmp_max_ram = None
            tmp_max_cpu = None

        # Put useful pieces into dictionary
        task_dict['Tool Name'] = tmp_name
        task_dict['Task ID'] = task_id
        task_dict['Exit Code'] = tmp_ecod
        task_dict['RAM: Max (MB)'] = tmp_max_ram
        task_dict['RAM: Series (MB)'] = tmp_ram
        task_dict['CPU: Max (%)'] = tmp_max_cpu
        task_dict['CPU: Series (%)'] = tmp_cpu
        task_dict['Time: Total (s)'] = tmp_durr
        task_dict['Time: Series (s)'] = tmp_tim
        task_dict['Time: Start'] = tmp_time
        task_dict.update(tmp_invo_dict)
        task_dict['Log: Output'] = tmp_sout
        task_dict['Log: Error'] = tmp_serr

        # Add task to the experiment
        experiment += [task_dict]

    # Save experiment to file
    with open(outfile, 'w') as fhandle:
        fhandle.write(json.dumps(experiment, indent=4, sort_keys=True))

    return experiment



#!/usr/bin/env python
#
# This software is distributed with the MIT license:
# https://github.com/gkiar/clowdr/blob/master/LICENSE
#
# clowdr/controller/launcher.py
# Created by Greg Kiar on 2018-02-28.
# Email: gkiar@mcin.ca


class Endpoint:
    # TODO: document

    def __init__(self, auth):
        # TODO: document
        self.credentials = auth



from setuptools import setup
import os.path as op

VERSION = "0.0.17"
DEPS = [
         "boto3",
         "flask",
         "boutiques>=0.5.10",
         "pandas>=0.23.4",
         "slurmpy==0.0.7",
         "dash>=0.24.1",
         "dash-renderer>=0.13.0",
         "dash-html-components>=0.11.0",
         "dash-core-components>=0.27.1",
         "dash-table-experiments>=0.6.0",
         "colorlover",
         "psutil",
         "plotly>=3.1.1"
       ]

setup(name="clowdr",
      version=VERSION,
      description="Tool for launching pipelines locally and remotely",
      url="http://github.com/clowdr/clowdr",
      author="Gregory Kiar",
      author_email="gkiar07@gmail.com",
      python_requires=">=3",
      classifiers=[
                "Programming Language :: Python",
                "Programming Language :: Python :: 3",
                "Programming Language :: Python :: 3.5",
                "Programming Language :: Python :: 3.6",
                "Programming Language :: Python :: 3.7",
                "Programming Language :: Python :: Implementation :: PyPy",
                "License :: OSI Approved :: MIT License",
                "Topic :: Software Development :: Libraries :: Python Modules",
                "Operating System :: OS Independent",
                "Natural Language :: English"
                  ],
      license="MIT",
      packages=["clowdr",
                "clowdr.controller",
                "clowdr.endpoint",
                "clowdr.share"],
      include_package_data=True,
      package_data = {"clowdr/templates": [op.join("templates", "AWS","*")]},
      test_suite="pytest",
      tests_require=["pytest", "pytest-runner", "pycodestyle"],
      setup_requires=DEPS,
      install_requires=DEPS,
      entry_points = {
        "console_scripts": [
            "clowdr=clowdr.driver:main",
        ]
      },
      zip_safe=False)



#!/usr/bin/env python
#
# This software is distributed with the MIT license:
# https://github.com/gkiar/clowdr/blob/master/LICENSE
#
# clowdr/task.py
# Created by Greg Kiar on 2018-02-28.
# Email: gkiar@mcin.ca

from argparse import ArgumentParser
from datetime import datetime
from time import mktime, localtime
from subprocess import PIPE
import multiprocessing as mp
import numpy as np
import os.path as op
import subprocess
import psutil
import time
import json
import csv
import os
import re
import warnings


warnings.filterwarnings("ignore", message="numpy.dtype size changed")

import pandas as pd

import boutiques as bosh
from clowdr import utils


class TaskHandler:
    def __init__(self, taskfile, **kwargs):
        self.manageTask(taskfile, **kwargs)

    def manageTask(self, taskfile, provdir=None, verbose=False, **kwargs):
        # Get metadata
        if provdir is None:
            self.localtaskdir = "/clowtask/"
        else:
            self.localtaskdir = provdir

        # The below grabs an ID from the form: /some/path/to/fname-#.ext
        self.task_id = taskfile.split('.')[0].split('-')[-1]

        self.localtaskdir = op.join(self.localtaskdir, "clowtask_"+self.task_id)
        if not op.exists(self.localtaskdir):
            os.makedirs(self.localtaskdir)

        if(verbose):
            print("Fetching metadata...", flush=True)
        remotetaskdir = op.dirname(taskfile)
        taskfile = utils.get(taskfile, self.localtaskdir)[0]

        # Parse metadata
        taskinfo = json.load(open(taskfile))
        descriptor = taskinfo['tool']
        invocation = taskinfo['invocation']
        input_data = taskinfo['dataloc']
        output_loc = utils.truepath(taskinfo['taskloc'])

        if(verbose):
            print("Fetching descriptor and invocation...", flush=True)
        # Get descriptor and invocation
        desc_local = utils.get(descriptor, self.localtaskdir)[0]
        invo_local = utils.get(invocation, self.localtaskdir)[0]

        # Get input data, if running remotely
        if not kwargs.get("local") and \
           any([dl.startswith("s3://") for dl in input_data]):
            if(verbose):
                print("Fetching input data...", flush=True)
            localdatadir = op.join("/data")
            local_input_data = []
            for dataloc in input_data:
                local_input_data += utils.get(dataloc, localdatadir)
            # Move to correct location
            os.chdir(localdatadir)
        else:
            if(verbose):
                print("Skipping data fetch (local execution)...", flush=True)
            if kwargs.get("workdir") and op.exists(kwargs.get("workdir")):
                os.chdir(kwargs["workdir"])

        if(verbose):
            print("Beginning execution...", flush=True)
        # Launch task
        copts = ['launch', desc_local, invo_local]
        if kwargs.get("volumes"):
            volumes = " ".join(kwargs.get("volumes"))
            copts += ['-v', volumes]
        if kwargs.get("user"):
            copts += ['-u']

        start_time = time.time()
        self.provLaunch(copts, verbose=verbose, **kwargs)
        if(verbose):
            print(self.output, flush=True)
        duration = time.time() - start_time

        # Get list of bosh exec outputs
        with open(desc_local) as fhandle:
            outputs_all = json.load(fhandle)["output-files"]

        outputs_present = []
        outputs_all = bosh.evaluate(desc_local, invo_local, 'output-files/')
        for outfile in outputs_all.values():
            outputs_present += [outfile] if op.exists(outfile) else []

        # Write memory/cpu stats to file
        usagef = "task-{}-usage.csv".format(self.task_id)
        self.cpu_ram_usage.to_csv(op.join(self.localtaskdir, usagef),
                                  sep=',', index=False)
        utils.post(op.join(self.localtaskdir, usagef), remotetaskdir)

        # Write stdout to file
        stdoutf = "task-{}-stdout.txt".format(self.task_id)
        with open(op.join(self.localtaskdir, stdoutf), "w") as fhandle:
            fhandle.write(self.output.stdout)
        utils.post(op.join(self.localtaskdir, stdoutf), remotetaskdir)

        # Write sterr to file
        stderrf = "task-{}-stderr.txt".format(self.task_id)
        with open(op.join(self.localtaskdir, stderrf), "w") as fhandle:
            fhandle.write(self.output.stderr)
        utils.post(op.join(self.localtaskdir, stderrf), remotetaskdir)

        start_time = datetime.fromtimestamp(mktime(localtime(start_time)))
        summary = {"duration": duration,
                   "launchtime": str(start_time),
                   "exitcode": self.output.exit_code,
                   "outputs": [],
                   "usage": op.join(remotetaskdir, usagef),
                   "stdout": op.join(remotetaskdir, stdoutf),
                   "stderr": op.join(remotetaskdir, stderrf)}

        if not kwargs.get("local"):
            if(verbose):
                print("Uploading outputs...", flush=True)
            # Push outputs
            for local_output in outputs_present:
                if(verbose):
                    print("{} --> {}".format(local_output, output_loc),
                          flush=True)
                tmpouts = utils.post(local_output, output_loc)
                print(tmpouts)
                summary["outputs"] += tmpouts
        else:
            if(verbose):
                print("Skipping uploading outputs (local execution)...",
                      flush=True)
            summary["outputs"] = outputs_present

        summarf = "task-{}-summary.json".format(self.task_id)
        with open(op.join(self.localtaskdir, summarf), "w") as fhandle:
            fhandle.write(json.dumps(summary, indent=4, sort_keys=True) + "\n")
        utils.post(op.join(self.localtaskdir, summarf), remotetaskdir)

        # If not local, delete all: inputs, outputs, and summaries
        if not kwargs.get("local"):
            for local_output in outputs_present:
                utils.remove(local_output)
            utils.remove(self.localtaskdir)
            for local_input in local_input_data:
                utils.remove(local_input)

    def execWrapper(self, sender):
        # if reprozip: use it
        if not subprocess.Popen("type reprozip 2>/dev/null", shell=True).wait():
            if self.runner_kwargs.get("verbose"):
                print("Reprozip found; will use to record provenance!",
                      flush=True)
            cmd = 'reprozip usage_report --disable'
            p = subprocess.Popen(cmd, shell=True).wait()

            cmd = 'reprozip trace -w --dir={}/task-{}-reprozip/ bosh exec {}'
            p = subprocess.Popen(cmd.format(self.localtaskdir,
                                            self.task_id,
                                            " ".join(self.runner_args)),
                                 shell=True).wait()

            cmd = ('reprozip pack --dir={0}/task-{1}-reprozip/ '
                   '{0}/task-{1}-reprozip'.format(self.localtaskdir,
                                                  self.task_id))
            p = subprocess.Popen(cmd, shell=True).wait()
        else:
            if self.runner_kwargs.get("verbose"):
                print("Reprozip not found; install to record more provenance!",
                      flush=True)
            sender.send(bosh.execute(*self.runner_args))

    def provLaunch(self, options, **kwargs):
        self.runner_args = options
        self.runner_kwargs = kwargs
        timing, cpu, ram = self.monitor(self.execWrapper, **kwargs)

        basetime = timing[0]

        total_df = pd.DataFrame(columns=['time', 'cpu', 'ram'])
        for ttime, tcpu, tram in zip(timing, cpu, ram):
            total_df.loc[len(total_df)] = (ttime-basetime, tcpu, tram)

        self.cpu_ram_usage = total_df

    def monitor(self, target, **kwargs):
        ram_lut = {'B': 1/1024/1024,
                   'KiB': 1/1024,
                   'MiB': 1,
                   'GiB': 1024}
        self.output, sender = mp.Pipe(False)
        worker_process = mp.Process(target=target, args=(sender,))
        worker_process.start()
        p = psutil.Process(worker_process.pid)

        log_time = []
        log_cpu = []
        log_mem = []
        while worker_process.is_alive():
            try:
                cpu = p.cpu_percent()
                ram = p.memory_info()[0]*ram_lut['B']

                for subproc in p.children(recursive=True):
                    if not subproc.is_running():
                        continue

                    subproc_dict = subproc.as_dict(attrs=['pid',
                                                          'name',
                                                          'cmdline',
                                                          'cpu_percent',
                                                          'memory_info'])
                    if subproc_dict['name'] == 'docker':
                        call = subproc_dict['cmdline'][-1]
                        tcmd = psutil.Popen(["docker", "ps", "-q"], stdout=PIPE)
                        running = tcmd.communicate()[0].decode('utf-8')
                        running = running.split('\n')
                        tcmd = psutil.Popen(["docker", "inspect"] + running,
                                            stdout=PIPE, stderr=PIPE)
                        tinf = json.loads(tcmd.communicate()[0].decode('utf-8'))
                        for tcon in tinf:
                            if (tcon.get("Config") and
                               tcon.get("Config").get("Cmd") and
                               call in tcon['Config']['Cmd']):
                                tid = tcon['Id']
                                tcmd = psutil.Popen([
                                                     "docker",
                                                     "stats",
                                                     tid,
                                                     "--no-stream",
                                                     "--format",
                                                     "'{{.MemUsage}} " +\
                                                     "{{.CPUPerc}}'"
                                                    ],
                                                    stdout=PIPE)
                                tout = tcmd.communicate()[0].decode('utf-8')
                                tout = tout.strip('\n').replace("'", "")

                                _ram, _, _, _cpu = tout.split(' ')
                                _ram, ending = re.match('([0-9.]+)([MGK]?i?B)',
                                                        _ram).groups()
                                ram += float(_ram) * ram_lut[ending]
                                cpu += float(_cpu.strip('%'))

                    else:
                        cpu += subproc_dict['cpu_percent']
                        ram += subproc_dict['memory_info'][0] * ram_lut['B']

                    if kwargs.get('verbose'):
                        print(cpu, ram)

                tim = time.time()
                log_time.append(tim)
                log_cpu.append(cpu)
                log_mem.append(ram)
                time.sleep(0.1)

            except (psutil._exceptions.AccessDenied,
                    psutil._exceptions.NoSuchProcess,
                    TypeError, ValueError, AttributeError) as e:
                continue

        worker_process.join()
        self.output = self.output.recv()
        return log_time, log_cpu, log_mem



#!/usr/bin/env python

from shutil import copy, copytree, rmtree, SameFileError
from subprocess import Popen, PIPE, CalledProcessError
import os.path as op
import random as rnd
import string
import boto3
import time
import csv
import sys
import os
import re


def backoff(function, posargs, optargs, backoff_time=36000, **kwargs):
    fib_lo = 0
    fib_hi = 1
    while True:
        try:
            value = function(*posargs, **optargs)
            return (0, value)
        except Exception as e:
            if kwargs.get("verbose"):
                print("Failed ({}). Retrying in: {}s".format(type(e).__name__,
                                                             fib_hi))
            if fib_hi > backoff_time:
                if kwargs.get("verbose"):
                    print("Failed. Skipping!")
                return (-1, str(e))
            time.sleep(fib_hi)
            fib_lo, fib_hi = fib_hi, fib_lo + fib_hi


def getContainer(savedir, container, **kwargs):
    if container["type"] == "singularity":
        name = container.get("image")
        local = name.replace("/", "-").replace(":", "-")
        index = container.get("index")
        if not index:
            index = "shub://"
        elif not index.endswith("://"):
            index = index + "://"
        if kwargs.get("simg"):
            return get(kwargs["simg"], local + ".simg")
        else:
            cmd = "singularity pull --name \"{}.simg\" {}{}".format(local,
                                                                    index,
                                                                    name)
            if kwargs.get("verbose"):
                print(cmd)
            p = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)
            stdout = p.communicate()
            if kwargs.get("verbose"):
                try:
                    print(stdout.decode('utf-8'))
                except Exception as e:
                    print(stdout)
            return stdout


def truepath(path):
    if path.startswith("s3://"):
        return path
    else:
        return op.realpath(path)


def randstring(k):
    return "".join([rnd.choice(string.ascii_uppercase + string.digits)
                    for _ in range(k)])


def splitS3Path(path):
    return re.match('^s3://([a-zA-Z0-9_-]+)/([a-zA-Z0-9_/.-]+)',
                    path).group(1, 2)


def get(remote, local, **kwargs):
    try:
        if remote.startswith("s3://"):
            return _awsget(remote, local)
        elif op.isdir(remote):
            return [op.realpath(copytree(remote, local))]
        else:
            return [op.realpath(copy(remote, local))]
    except SameFileError as e:
        if kwargs.get("verbose"):
            print("SameFileWarning: some files may not have been moved")
        if op.isdir(local) and op.isfile(remote):
            return [op.realpath(op.join(local, op.basename(remote)))]
        else:
            return [op.realpath(local)]
    except FileExistsError as e:
        if kwargs.get("verbose"):
            print("FileExistsWarning: some files may not have been moved")
        if op.isdir(local) and op.isfile(remote):
            return [op.realpath(op.join(local, op.basename(remote)))]
        else:
            return [op.realpath(local)]


def post(local, remote, **kwargs):
    try:
        if remote.startswith("s3://"):
            return _awspost(local, remote)
        elif op.isdir(local):
            return [op.realpath(copytree(local, remote))]
        elif op.isdir(remote):
            return [op.realpath(copy(local,
                                     op.join(remote, op.basename(local))))]
        else:
            return [op.realpath(copy(local, remote))]
    except SameFileError as e:
        if kwargs.get("verbose"):
            print("SameFileWarning: some files may not have been moved")
        if op.isdir(remote) and op.isfile(local):
            return [op.realpath(op.join(remote, op.basename(local)))]
        else:
            return [op.realpath(remote)]


def remove(local):
    try:
        if op.isfile(local):
            os.remove(local)
        elif op.isdir(local):
            rmtree(local)
    except FileNotFoundError as e:
        pass


def _awsget(remote, local):
    s3 = boto3.resource("s3")

    bucket, rpath = splitS3Path(remote)

    buck = s3.Bucket(bucket)
    files = [obj.key
             for obj in buck.objects.filter(Prefix=rpath)
             if not os.path.isdir(obj.key)]
    files_local = []
    for fl in files:
        fl_local = op.join(local, fl)
        files_local += [fl_local]
        os.makedirs(op.dirname(fl_local), exist_ok=True)
        if fl_local.strip('/') == op.dirname(fl_local).strip('/'):
            continue  # Create, but don't try to download directories
        buck.download_file(fl, fl_local)

    return files_local


def _awspost(local, remote):
    # Credit: https://github.com/boto/boto3/issues/358#issuecomment-346093506
    if not op.isfile(local):
        local_files = [op.join(root, f)
                       for root, dirs, files in os.walk(local)
                       for f in files]
    else:
        local_files = [local]

    s3 = boto3.client("s3")
    bucket, rpath = splitS3Path(remote)

    rempats = []
    for flocal in local_files:
        if local == flocal:
            rempat = op.join(rpath, op.basename(flocal))
        else:
            rempat = op.join(rpath, op.relpath(flocal, local))

        s3.upload_file(flocal, bucket, rempat, {'ACL': 'public-read'})
        rempats += [op.join('s3://', bucket, rempat)]

    return rempats



from __future__ import absolute_import

from .task import TaskHandler
from .controller import metadata, launcher, rerunner
from .endpoint import AWS, remote
from .driver import local, cloud, share
from .share import consolidate, portal, customDash

__all__ = ['task', 'controller', 'driver', 'endpoint', 'share']
